#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include <map>
#include <vector>
#include <string>
#include <algorithm>

using namespace std;

#define PRIOR_MU 1500
#define PRIOR_WEIGHT 1.0
#define MAX_PLAYERS 4096
#define DUMP_RAW 0

float mu[MAX_PLAYERS];
float mu_stddev[MAX_PLAYERS];
float global_sigma = 70.0f;
float prior_sigma = 70.0f;

#define EPSILON 1e-3

/*
 * L(mu_vec, sigma_vec, matches) = product[ L(mu_A, sigma_A, mu_B, sigma_B, score_AB - score_BA) ]
 * log-likelihood = sum[ log( L(mu_A, sigma_A, mu_B, sigma_B, score_AB - score_BA) ) ]
 * 
 * L(mu1, sigma1, mu2, sigma2, score2 - score1) = sigmoid(mu2 - mu1, sqrt(sigma1² + sigma2²), (score2 - score1))
 *
 * pdf := 1/(sigma * sqrt(2*Pi)) * exp(-(x - mu)^2 / (2 * sigma^2));        
 * pdfs := subs({ mu = mu1 - mu2, sigma = sqrt(sigma1^2 + sigma2^2) }, pdf);
 * diff(log(pdfs), mu1); 
 */

struct match {
	int player, other_player;
	int margin;
	float weight;
};
map<int, vector<match> > matches_for_player;
vector<match> all_matches;

void dump_scores(const vector<string> &players, const float *mu, const float *mu_stddev, int num_players)
{
#if 0
	for (int i = 0; i < num_players; ++i) {
		printf("%s=[%5.1f, %4.1f] ", players[i].c_str(), mu[i], sigma[i]);
	}
	printf("\n");
#elif 0
	for (int i = 0; i < num_players; ++i) {
		printf("%5.1f ", mu[i]);
	}
	printf("\n");
#else
	for (int i = 0; i < num_players; ++i) {
		printf("%f %f %s\n", mu[i], mu_stddev[i], players[i].c_str());
	}
#endif
}

/*
 * diff(logL, mu1) = -w * (mu1 - mu2 - x) / sigma_c^2
 * maximizer for mu1 is given by: sum_i[ (w_i/sigma_c_i)^2 (mu1 - mu2_i - x_i) ] = 0
 *                                sum_i[ (w_i/sigma_c_i)^2 mu1 ] = sum_i [ (w_i/sigma_c_i)^2 ( mu2_i + x_i ) ]
 *                                mu1 = sum_i [ (w_i/sigma_c_i)^2 ( mu2_i + x_i ) ] / sum_i[ (w_i/sigma_c_i)^2 ]
 */
void update_mu(float *mu, int player_num, const vector<match> &matches)
{
	if (matches.empty()) {
		return;
	}

	float nom = 0.0f, denom = 0.0f;

	// Prior.
	{
		float inv_sigma2 = 1.0f / (prior_sigma * prior_sigma);
		nom += PRIOR_WEIGHT * PRIOR_MU * inv_sigma2;
		denom += PRIOR_WEIGHT * inv_sigma2;
	}

	// All matches.
	for (unsigned i = 0; i < matches.size(); ++i) {
		float inv_sigma_c2 = matches[i].weight / (global_sigma * global_sigma);
		float x = matches[i].margin; // / 70.0f;
	
		nom += (mu[matches[i].other_player] + x) * inv_sigma_c2;
		denom += inv_sigma_c2;
	}
	mu[player_num] = nom / denom;
}

void dump_raw(const float *mu, int num_players)
{
	for (unsigned i = 0; i < all_matches.size(); ++i) {
		const match& m = all_matches[i];

		float mu1 = mu[m.player];
		float mu2 = mu[m.other_player];
		float sigma = global_sigma;
		float mu = mu1 - mu2;
		float x = m.margin;
		float w = m.weight;

		printf("%f %f\n", (x - mu) / sigma, w);
	}
}

/*
 * diff(logL, sigma) = w ( (x - mu)² - sigma² ) / sigma³
 * maximizer for sigma is given by: sum_i[ (w_i/sigma)³ ((x - mu)² - sigma²) ] = 0
 *                                   sum_i[ w_i ( (x - mu)² - sigma² ) ] = 0                            |: sigma != 0
 *                                   sum_i[ w_i (x - mu)² ] = sum[ w_i sigma² ]
 *                                   sigma = sqrt( sum_i[ w_i (x - mu)² ] / sum[w_i] )
 */
void update_global_sigma(float *mu, int num_players)
{
	float nom = 0.0f, denom = 0.0f;
	for (unsigned i = 0; i < all_matches.size(); ++i) {
		const match& m = all_matches[i];

		float mu1 = mu[m.player];
		float mu2 = mu[m.other_player];
		float mu = mu1 - mu2;
		float x = m.margin;
		float w = m.weight;

		nom += w * ((x - mu) * (x - mu));
		denom += w;
	}

	global_sigma = sqrt(nom / denom);
}

/*
 * diff(priorlogL, sigma) = w ( (x - mu)² - sigma² ) / sigma³
 * maximizer for sigma is given by: sum_i[ (w_i/sigma)³ ((x - mu)² - sigma²) ] = 0
 *                                   sum_i[ w_i ( (x - mu)² - sigma² ) ] = 0                            |: sigma != 0
 *                                   sum_i[ w_i (x - mu)² ] = sum[ w_i sigma² ]
 *                                   sigma = sqrt( sum_i[ w_i (x - mu)² ] / sum[w_i] )
 */
void update_prior_sigma(float *mu, int num_players)
{
	float nom = 0.0f, denom = 0.0f;
	for (int i = 0; i < num_players; ++i) {
		float mu1 = mu[i];

		nom += ((mu1 - PRIOR_MU) * (mu1 - PRIOR_MU));
		denom += 1.0f;
	}

	prior_sigma = sqrt(nom / denom);
	if (!(prior_sigma > 40.0f)) {
		prior_sigma = 40.0f;
	}
}

float compute_logl(float z)
{
	return -0.5 * (log(2.0f / M_PI) + z * z);
}

float compute_total_logl(float *mu, int num_players)
{
	float total_logl = 0.0f;

	// Prior.
	for (int i = 0; i < num_players; ++i) {
		total_logl += PRIOR_WEIGHT * compute_logl((mu[i] - PRIOR_MU) / prior_sigma);
	}

	// Matches.
	for (unsigned i = 0; i < all_matches.size(); ++i) {
		const match& m = all_matches[i];

		float mu1 = mu[m.player];
		float mu2 = mu[m.other_player];
		float sigma = global_sigma;
		float mu = mu1 - mu2;
		float x = m.margin;
		float w = m.weight;

		total_logl += w * compute_logl((x - mu) / sigma);
	}

	return total_logl;
}

/*
 * Compute Hessian matrix of the negative log-likelihood, ie. for each term in logL:
 *
 * M_ij = D_i D_j (- logL) = -w / sigma²                                for i != j
 *                            w / sigma²                                for i == j
 *
 * Note that this does not depend on mu or the margin at all.
 */
double hessian[MAX_PLAYERS][MAX_PLAYERS];
void construct_hessian(const float *mu, int num_players)
{
	memset(hessian, 0, sizeof(hessian));

	for (unsigned i = 0; i < all_matches.size(); ++i) {
		const match &m = all_matches[i];

		int p1 = m.player;
		int p2 = m.other_player;

		double sigma_sq = global_sigma * global_sigma;
		float w = m.weight;

		hessian[p1][p2] -= w / sigma_sq;
		hessian[p2][p1] -= w / sigma_sq;

		hessian[p1][p1] += w / sigma_sq;
		hessian[p2][p2] += w / sigma_sq;
	}
}

// Compute uncertainty (stddev) of mu estimates, which is 1/sqrt(H_ii),
// where H is the Hessian (see construct_hessian()).
void compute_mu_uncertainty(const float *mu, int num_players)
{
	memset(mu_stddev, 0, sizeof(mu_stddev));

	for (unsigned i = 0; i < all_matches.size(); ++i) {
		const match &m = all_matches[i];

		int p1 = m.player;
		int p2 = m.other_player;

		double sigma_sq = global_sigma * global_sigma;
		float w = m.weight;

		// Temporarily use mu_stddev to store the diagonal of the Hessian.
		mu_stddev[p1] += w / sigma_sq;
		mu_stddev[p2] += w / sigma_sq;
	}
	for (int i = 0; i < num_players; ++i) {
		mu_stddev[i] = 1.0f / sqrt(mu_stddev[i]);
	}
}

int main(int argc, char **argv)
{
	int num_players;
	if (scanf("%d", &num_players) != 1) {
		fprintf(stderr, "Could't read number of players\n");
		exit(1);
	}

	if (num_players > MAX_PLAYERS) {
		fprintf(stderr, "Max %d players supported\n", MAX_PLAYERS);
		exit(1);
	}

	vector<string> players;
	map<string, int> player_map;

	for (int i = 0; i < num_players; ++i) {
		char buf[256];
		if (scanf("%s", buf) != 1) {
			fprintf(stderr, "Couldn't read player %d\n", i);
			exit(1);
		}

		players.push_back(buf);
		player_map[buf] = i;
	}

	int num_matches = 0;
	for ( ;; ) {
		char pl1[256], pl2[256];
		int score1, score2;
		float weight;

		if (scanf("%s %s %d %d %f", pl1, pl2, &score1, &score2, &weight) != 5) {
			//fprintf(stderr, "Read %d matches.\n", num_matches);
			break;
		}

		++num_matches;

		if (player_map.count(pl1) == 0) {
			fprintf(stderr, "Unknown player '%s'\n", pl1);
			exit(1);
		}
		if (player_map.count(pl2) == 0) {
			fprintf(stderr, "Unknown player '%s'\n", pl2);
			exit(1);
		}

		match m1;
		m1.player = player_map[pl1];
		m1.other_player = player_map[pl2];
		m1.margin = score1 - score2;
		m1.weight = weight;
		matches_for_player[player_map[pl1]].push_back(m1);

		match m2;
		m2.player = player_map[pl2];
		m2.other_player = player_map[pl1];
		m2.margin = score2 - score1;
		m2.weight = weight;
		matches_for_player[player_map[pl2]].push_back(m2);

		all_matches.push_back(m1);
	}

	float mu[MAX_PLAYERS];

	for (int i = 0; i < num_players; ++i) {
		mu[i] = PRIOR_MU;
	}

	for (int j = 0; j < 1000; ++j) {
		float old_mu[MAX_PLAYERS];
		float old_global_sigma = global_sigma;
		float old_prior_sigma = prior_sigma;
		memcpy(old_mu, mu, sizeof(mu));
		for (int i = 0; i < num_players; ++i) {
			update_mu(mu, i, matches_for_player[i]);
		}
		update_global_sigma(mu, num_players);
		update_prior_sigma(mu, num_players);
		/* for (int i = 0; i < num_players; ++i) {
			update_sigma(mu, sigma, i, matches_for_player[i]);
			dump_scores(players, mu, sigma, num_players);
		} */

		float sumdiff = 0.0f;
		for (int i = 0; i < num_players; ++i) {
			sumdiff += (mu[i] - old_mu[i]) * (mu[i] - old_mu[i]);
		}
		sumdiff += (prior_sigma - old_prior_sigma) * (prior_sigma - old_prior_sigma);
		sumdiff += (global_sigma - old_global_sigma) * (global_sigma - old_global_sigma);
		if (sumdiff < EPSILON) {
			//fprintf(stderr, "Converged after %d iterations. Stopping.\n", j);
			printf("%d -1\n", j + 1);
			break;
		}
	}

#if DUMP_RAW
	dump_raw(mu, num_players);
#else
	compute_mu_uncertainty(mu, num_players);
	dump_scores(players, mu, mu_stddev, num_players);
	//fprintf(stderr, "Optimal sigma: %f (two-player: %f)\n", sigma[0], sigma[0] * sqrt(2.0f));
	printf("%f -2\n", global_sigma / sqrt(2.0f));
	printf("%f -3\n", prior_sigma);

	float total_logl = compute_total_logl(mu, num_players);
	printf("%f -4\n", total_logl);

//	construct_hessian(mu, sigma, num_players);
#endif
}
